/*
	address.c -- recipient handling code
	(C) Wouter Coene, 1999

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, write to the Free Software
	Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "config-wrap.h"

#include "address.h"
#include "mailer.h"
#include "lib.h"
#include "log.h"

/***
 * format_email
 * Format an email address according to rules specified in
 * RFC #822.
 */
static char *format_email(const char *str)
{
	char *s, *p, *t, *n;
	int in_brackets;

	s=strdup(str);
	t=stralloc(strlen(str));

	in_brackets=FALSE;

	p=s;
	while (*p!='\0') {
		if (strchr(EMAIL_LITERAL,lower(*p))!=NULL) {
			strncat(t,p,1);
		} else if (*p=='"') {
			if ((n=strchr(p+1,'"'))!=NULL) {
				strncat(t,p,(n-p)+1);
				p=n;
			}
		} else if (*p=='(') {
			if ((n=strchr(p+1,')'))!=NULL)
				p=n;
		} else if (*p=='<') {
			*t='\0';
			in_brackets=TRUE;
		} else if (*p=='>') {
			if (in_brackets)
				break;
		}

		p++;
	}

	free(s);
	s=strdup(t);
	free(t);

	return s;
}

/***
 * address_alloc
 * Allocate a t_address struct and optionally set the to field.
 */
t_address *address_alloc(const char *to)
{
	t_address *address;

	address=calloc(sizeof(t_address),1);
	if (to!=NULL)
		address->addr=strdup(to);

	return address;
}

/***
 * address_free
 * Free a previously allocated t_address struct.
 */
t_address *address_free(t_address *address)
{
	if (address==NULL)
		return NULL;

	if (address->addr!=NULL)
		free(address->addr);
	if (address->formatted!=NULL)
		free(address->formatted);
	if (address->user!=NULL)
		free(address->user);
	if (address->domain!=NULL)
		free(address->domain);
	free(address);

	return NULL;
}

/***
 * address_format
 * Format the recipient address in a t_address struct.
 */
t_address *address_format(t_address *address)
{
	char *s, *p;

	log_enter();
	if ((address==NULL) || (address->addr==NULL))
		return address;

	if (address->formatted!=NULL)
		free(address->formatted);
	s=strdup(address->formatted=format_email(address->addr));

	p=strchr(s,'@');
	if (p!=NULL) {
		*p='\0';
		p++;
	}

	if (address->user!=NULL)
		free(address->user);
	address->user=strdup(s);
	if (address->domain!=NULL)
		free(address->domain);
	if (p!=NULL)
		address->domain=strdup(p);
	else
		address->domain=NULL;

	free(s);
	log_leave();

	return address;
}

/***
 * address_islocal
 * Check if a given recipient is a local recipient.
 */
int address_islocal(t_address *address, char *locals[], int local_count)
{
	int i;

	log_enter();
	if ((address==NULL) || (address->addr==NULL))
		return FALSE;

	if (address->formatted==NULL)
		address=address_format(address);

	if (address->domain==NULL)
		return TRUE;

	for (i=0; i<local_count; i++) {
		if (strcasecmp(address->domain,locals[i])==0)
			return TRUE;
	}
	log_leave();

	return FALSE;
}

/***
 * address_count
 * Return the number of recipients this t_address struct points to
 * (follow the linked list).
 */
int address_count(t_address *address_start)
{
	t_address *address;
	int count;

	count=0;
	address=address_start;
	while (address!=NULL) {
		count++;
		address=address->next;
	}

	return count;
}

/* end address.c */
