/*
	cmdline.c -- command line parser for OMTA
	(C) Wouter Coene, 1999

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, write to the Free Software
	Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define WANT_UNISTD
#include "config-wrap.h"
#include <getopt.h>

#define CMDLINE_HAVE_DATA
#include "cmdline.h"
#include "omta.h"
#include "mailer.h"
#include "address.h"
#include "readconf.h"
#include "lib.h"
#include "log.h"

char *myself=NULL;
t_address *from=NULL, *to=NULL;

/***
 * help
 * Write a usage message to the standard output.
 */
void help()
{
	printf("Usage:\t%s [mode] [options] [address...]\n\n",myself);
	printf(
		"     -bd, --daemon\tdaemon mode\n"
		"     -bD\t\tnon-forked daemon mode\n"
		"     -bs, --smtp\tSMTP over stdin\n"
		"     -bm, --plain\tplain mode (read msg from stdin)\n"
		"     -bp, --dump\tqueue print mode\n"
		"     -q,  --transfer\tqueue run mode\n\n"

		"     -t\t\t\tscan message for recipients\n"
		"     -i,  --ignore-dots\tignore dots on lines by themselves\n"
		"     -C,  --config\tspecify config file\n\n"

		"     -f,  --from\tset the address of the sender\n"
		"     -F,  --from-name\tset the name of the sender\n\n"

		"     -h,  --help\tshow this information\n"
		"     -v,  --version\toutput version and license info\n"
		"     -D,  --default\tshow compile-time defaults\n\n"

		"     -d,  --log-dialog\tlog any SMTP dialogs\n"
		"     -s,  --silent\tdon't log anything at all\n"
		"     -l,  --log-level\tset logging level (see man page)\n\n"

		"Please report bugs to <" OMTA_LIST ">\n");
}

/***
 * short_help
 * Write a GNU-style short help message to the standard output.
 */
void short_help()
{
	printf("Try `%s --help' for more information.\n",myself);
}

/***
 * short_version
 * Write a single-line version string to the standard output.
 */
void short_version()
{
	printf(NAME " " VERSION ", by the OMTA people <" OMTA_LIST ">\n");
}

/***
 * version
 * Write the full version number and license information to the
 * standard output.
 */
void version()
{
	short_version();
	printf(
		"Get it at " OMTA_URI " \n\n"

		"This program is free software; you can redistribute it and/or modify\n"
		"it under the terms of the GNU General Public License as published by\n"
		"the Free Software Foundation; either version 2 of the License, or\n"
		"(at your option) any later version.\n\n"

		"This program is distributed in the hope that it will be useful,\n"
		"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
		"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
		"GNU General Public License for more details.\n\n"

		"You should have received a copy of the GNU General Public License\n"
		"along with this program. If not, write to the Free Software\n"
		"Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.\n");
}

/***
 * cmdline
 * Parse the command line.
 */
int cmdline(int argc, char *argv[])
{
	int opt, option_index=0, ok, i;
	t_address *address, *end_to=NULL;
	char *from_name=NULL, *from_email=NULL, *s;

	log_enter();
	myself=strdup(argv[0]);

	mode=MODE_PLAIN;
 	if (strstr(myself,"transfermail")==(myself+strlen(myself))-12) {
		mode=MODE_XFER;
	} else if (strstr(myself,"mailq")==(myself+strlen(myself))-5) {
		mode=MODE_DUMP;
	} else if (strstr(myself,"in.smtpd")==(myself+strlen(myself))-8) {
		mode=MODE_SMTP;
	} else {
		mode=MODE_PLAIN;
	}

	while ((opt=getopt_long(argc, argv, SHORT_OPTIONS, long_options, &option_index))!=-1) {
		ok=FALSE;
		switch(opt) {
			case 'h':
				mode=MODE_HELP;
				break;

			case 'v':
				mode=MODE_VERSION;
				break;

			case 'b':
				if (strlen(optarg)==1) {
					ok=TRUE;
					switch (*optarg) {
						case 'd':
							mode=MODE_DAEMON;
							do_fork=TRUE;
							break;

						case 'D':
							mode=MODE_DAEMON;
							break;

						case 's':
							mode=MODE_SMTP;
							break;

						case 'm':
							mode=MODE_PLAIN;
							break;

						case 'p':
							mode=MODE_DUMP;
							break;

						default:
							ok=FALSE;
					}
				}
				if (!ok) {
					fprintf(stderr,"%s: invalid argument for option b -- %s\n",myself,optarg);
					mode=MODE_SHORTHELP;
				}
				break;

			case 'i':
				ignore_period=TRUE;
				break;

			case 't':
				scan_message=TRUE;
				break;

			case 'q':
				if (optarg==NULL) { /* fix to disregard -q60m */
					mode=MODE_XFER;
				} else if (mode==MODE_PLAIN) {
					/* don't drop into mail send mode */
					mode=MODE_SHORTHELP;
				} else {
					fprintf(stderr,"%s: periodic queue flushing not implemented\n",myself);
					log_debug("%s: periodic queue flushing not implemented\n",myself);
					/* ...and will probably never get implemented ;)) */
				}
				break;

			case 'o':
				if (strcmp(optarg,"i")==0) {
					ignore_period=TRUE;
				} else if (strcmp(optarg,"em")==0) {
					/* NOP */
				} else {
					log_error("Ignoring invalid argument for option o -- %s\n",optarg);
				}
				break;

			case 'C':
				if (conf_global!=NULL)
					free(conf_global);
				conf_global=strdup(optarg);
				if (!file_exists(conf_global)) {
					perror(conf_global);
					mode=MODE_SHORTHELP;
				}
				break;				                                                                                                        

			case 'D':
				mode=MODE_DEFAULTS;
				break;

			case 'f':
				if (from_email!=NULL)
					free(from_email);
				from_email=strdup(optarg);
				break;

			case 'F':
				if (from_name!=NULL)
					free(from_name);
				from_name=strdup(optarg);
				break;

			case 'd':
				log_level|=OLOG_DIALOG;
				break;

			case 's':
				log_level=OLOG_NONE;
				break;

			case 'l':
				log_level=atoi(optarg);
				break;

			case '?':
				mode=MODE_SHORTHELP;
				break;
		}
		if (mode>=MODE_NOMODE)
			break;
	}

	while (optind<argc) {
		address=address_alloc(argv[optind++]);
		if (end_to==NULL) {
			to=address;
			end_to=address;
		} else {
			end_to->next=address;
			address->prev=end_to;
			end_to=address;
		}
	}

	if (from_email!=NULL) {
		i=strlen(from_email);
		if (from_name!=NULL)
			i+=strlen(from_name)+3;

		s=stralloc(i);
		if (from_name==NULL)
			snprintf(s,i+1,"%s",from_email);
		else
			snprintf(s,i+1,"%s <%s>",from_name,from_email);

		from=address_alloc(s);
		free(s);
	}

	log_leave();

	return mode;
}

/* end cmdline.c */
