/*
	daemon.c -- daemon mode routines
	(C) Wouter Coene, 1999

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, write to the Free Software
	Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define WANT_UNISTD
#define WANT_SYS_TYPES
#define WANT_TIME
#include "config-wrap.h"
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
                     
#include "daemon.h"
#include "smtpserv.h"
#include "readconf.h"
#include "log.h"
#include "omta.h"
#include "lib.h"

static int configured;

/***
 * sig_hup
 * HUP signal handler.
 */
static RETSIGTYPE sig_hup(int dummy)
{
	configured=FALSE;
}

/***
 * daemon_setup
 * Setup the daemon mode.
 */
static void daemon_setup()
{
	signal(SIGHUP,sig_hup);
	signal(SIGCHLD,SIG_IGN);
}

/***
 * socket_interface
 * Get the interface to be used for the socket.
 */
static struct in_addr socket_interface(const char *interface)
{
	struct in_addr in;
	struct hostent *h;

	in.s_addr=INADDR_ANY;
	if (interface==NULL)
		return in;

	h=gethostbyname(interface);
	if (h==NULL) {
		log_debug("socket_interface: %s\n",hstrerror(h_errno));
		log_error("Unable to resolve host %s.\n",interface);
		exit(2);
	}

	memcpy(&in,h->h_addr,h->h_length);

	return in;
}

/***
 * socket_open
 * Open a server socket at the specified port.
 */
static int socket_open(int port)
{
	struct sockaddr_in addr;
	int sid;

	if ((sid=socket(PF_INET,SOCK_STREAM,IPPROTO_IP))==-1) {
		log_error("socket_open (socket): %s\n",strerror(errno));
		return -1;
	}

	addr.sin_family=AF_INET;
	addr.sin_port=htons(port);
	addr.sin_addr=socket_interface(conf.interface);
	if (((conf.insecure_port) && (port<1024))
			|| (bind(sid,(struct sockaddr *) &addr,sizeof(addr))!=0)) {
		log_error("socket_open (bind): %s\n",strerror(errno));
		return -1;
	}

	if (listen(sid,5)!=0) {
		log_error("socket_open (listen): %s\n",strerror(errno));
		return -1;
	}

	return sid;
}

/***
 * handle_connect
 * Handle an incoming connection.
 */
static void handle_connect(int sid, struct sockaddr_in *addr)
{
	FILE *fd;

	if (sid==-1) {
		log_debug("handle_connect (accept): %s\n",strerror(errno));
		return;
	}

	switch (fork()) {
		case -1:
			log_error("handle_connect (fork): %s\n",strerror(errno));
			break;

		case 0: /* business as usual :) */
			break;

		default:
			close(sid);
			return;
	}

	if ((fd=fdopen(sid,"r+"))==NULL)
		return;

	smtp_server(fd,fd,conf.smtp_greeting,addr->sin_addr.s_addr);

	fflush(fd);
	shutdown(sid,0);
	fclose(fd);

	exit(0);
}

/***
 * mode_daemon
 * The daemon mode main routine.
 */
int mode_daemon()
{
	int ready;
	struct sockaddr_in addr;
	unsigned int addr_size,sid;
	fd_set rfds;
	struct timeval tv;
	int current_port;
	int avail;

	if (do_fork) {
		switch (fork()) {
			case -1:
				log_error("mode_daemon (fork): %s\n",strerror(errno));
				return TRUE;

			case 0: /* business as usual :) */
				break;

			default:
				return TRUE;
		}
	}

	setsid();
	chdir("/");

	fclose(stdin);
	fclose(stdout);

	current_port=0;
	configured=TRUE;
	ready=FALSE;
	daemon_setup();
	sid=0; /* to make gcc stop whining */
	addr_size=sizeof(addr);
	forever {
		if (!configured) {
			if ((ready) && (current_port!=conf.server_port)) {
				shutdown(sid,0);
				close(sid);
			}
			config(TRUE);
			ready=FALSE;
			configured=TRUE;
		}
		if (!ready) {
			if (conf.server_port==0) {
				log_error("No server port defined!\n");
				return FALSE;
			}
			if (current_port!=conf.server_port) {
				if ((sid=socket_open(conf.server_port))==-1)
					return FALSE;
				current_port=conf.server_port;
			}
			ready=TRUE;
#ifdef DEBUG
			if (do_fork)
#endif /* DEBUG */
			log_to_stderr=OLOG_NONE;
		}

		FD_ZERO(&rfds);
		FD_SET(sid,&rfds);

		tv.tv_sec=5;
		tv.tv_usec=0;

		avail=select(sid+1,&rfds,NULL,NULL,&tv);

		if (avail>0) {
			handle_connect(accept(sid,
				(struct sockaddr *) &addr,&addr_size),&addr);
		}
	}

	return FALSE;
}

/* end daemon.c */
