/*
	execute.c -- simple (and secure) execute routines
	(C) Wouter Coene, 1999

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, write to the Free Software
	Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define WANT_UNISTD
#define WANT_SYS_TYPES
#define WANT_WAIT
#include "config-wrap.h"

#include "execute.h"
#include "lib.h"

/***
 * make_argv
 * Messy routine to build an array from a command line while
 * taking quoting into account.
 */
static char **make_argv(const char *str)
{
	char *s, *p, *t;
	char **argv;
	int i;

	p=s=strdup(str);

	/* at first we only count */
	i=2;
	while (*p!='\0') {
		if (strchr(" \t",*p)!=NULL) {
			while (strchr(" \t",*p)!=NULL) {
				if (p=='\0')
					break;
				p++;
			}
			i++;
			continue;
		} else if (strchr("'\"",*p)!=NULL) {
			if ((t=strchr(p+1,*p))!=NULL)
				p=t+1;
			continue;
		}
		p++;
	}

	argv=calloc(sizeof(void *)*i,1);

	/* then we go for it */
	i=0;
	p=s;
	argv[i]=p;
	while (*p!='\0') {
		if (strchr(" \t",*p)!=NULL) {
			*p='\0';
			while (strchr(" \t",*p)!=NULL) {
				if (p=='\0')
					break;
				p++;
			}
			i++;
			argv[i]=p;
			continue;
		} else if (strchr("'\"",*p)!=NULL) {
			if ((t=strchr(p+1,*p))!=NULL) {
				memmove(p,p+1,strlen(p+1)+1);
				p=t-1;
				memmove(p,p+1,strlen(p+1)+1);
			}
			continue;
		}
		p++;
	}
	argv[i+1]=NULL;

	return argv;
}

/***
 * make_env
 * Build a suid-root safe environment for use with execute()
 * (only $PATH actually :^) ).
 */
static char **make_env()
{
	char **env;

	env=calloc(sizeof(void *)*2,1);
	env[0]=strdup("PATH=" SAFE_PATH);
	env[1]=NULL;

	return env;
}

/***
 * execute
 * Execute a command with an (almost empty) environment. This is
 * much safer if you're running suid-root.
 */
int execute(const char *path, int do_fork, int drop_priv)
{
	pid_t pid;
	int status;
	char **argv, **env, *cmd, *t;

	if (do_fork) {
		switch (pid=fork()) {
			case -1:
				perror("fork");
				return -1;

			case 0: /* bussiness as usual :) */
				break;

			default:
				waitpid(pid,&status,0);
				if (WIFEXITED(status))
					return WEXITSTATUS(status);
				else
					return -2;
				break;
		}
	}

	setsid();
	chdir("/");

	if (drop_priv) {
		setgid(gid);
		setuid(uid);
	}

	argv=make_argv(path);
	cmd=argv[0];
	if ((t=strrchr(argv[0],'/'))!=NULL) {
		argv[0]=t+1;
	}
	env=make_env();

	exit(execve(cmd,argv,env));
}

/***
 * pexecute
 * Safe popen() replacement, using execute() for the final
 * executing code.
 */
FILE *pexecute(const char *path, char mode, int drop_priv, int *saved_pid)
{
	int fd[2];

	if (strchr("rw",mode)==NULL) {
		errno=EINVAL;
		perror("pexecute");
		return NULL;
	}

	if (pipe(fd)!=0) {
		perror("pipe");
		return NULL;
	}

	switch ((*saved_pid)=fork()) {
		case -1:
			perror("fork");
			close(fd[0]);
			close(fd[1]);
			return NULL;

		case 0: /* bussiness as usual :) */
			break;

		default:
			usleep(50);
			if (waitpid(*saved_pid,NULL,WNOHANG)>0)
				return NULL;
			if (mode=='r') {
				close(fd[1]);
				return fdopen(fd[0],"r");
			} else {
				close(fd[0]);
				return fdopen(fd[1],"w");
			}
			break;
	}

	if (mode=='w') {
		close(fd[1]);
		dup2(fd[0],STDIN_FILENO);
		close(fd[0]);
	} else {
		close(fd[0]);
		dup2(fd[1],STDOUT_FILENO);
		close(fd[1]);
	}

	exit(execute(path,FALSE,drop_priv));
}

/***
 * pdone
 * Safe pclose() replacement, using execute() for the final
 * executing code.
 */
int pdone(FILE *fd, int saved_pid)
{
	int status;

	fclose(fd);

	waitpid(saved_pid,&status,0);
	if (WIFEXITED(status))
		return WEXITSTATUS(status);
	else
		return -2;
}

/* end execute.c */
