/*
	lib.c -- generic library routines
	(C) Wouter Coene, 1999

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, write to the Free Software
	Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define WANT_UNISTD
#define WANT_WAIT
#include "config-wrap.h"
#include <arpa/telnet.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/stat.h>
#include <pwd.h>

#include "lib.h"
#include "mailer.h"
#include "canon-host.h"
#include "log.h"

char *user=NULL, *user_gecos=NULL, *user_home=NULL;
char *fqdn=NULL, *node=NULL, *domain=NULL;

int uid, euid;
int gid, egid;

/***
 * lib_init
 * Library initialization function.
 */
void lib_init()
{
	struct passwd *pass;
	char *s, *t;
	struct stat st;

	uid=getuid();
	euid=geteuid();

	gid=getgid();
	egid=getegid();

	pass=getpwuid(uid);
	if ((pass!=NULL) && (pass->pw_name!=NULL)) {
		user=strdup(pass->pw_name);

		if ((t=strchr(pass->pw_gecos,','))!=NULL)
			*t='\0';
		user_gecos=strdup(pass->pw_gecos);
		user_home=strdup(pass->pw_dir);
	}

	s=stralloc(1024);
	if (gethostname(s,1024)<0) {
		log_error("Unable to get by own hostname!\n");
		exit(1);
	}
	if ((t=canon_host(s))!=NULL) {
		set_fqdn(t);
	} else {
		log_error("Unable to fully-qualify my hostname, using \"%s\"\n",s);
		set_fqdn(s);
	}
	free(s);

	fstat(fileno(stderr),&st);
	if (!S_ISCHR(st.st_mode)) {
		log_to_stderr=OLOG_NONE;
	}
}

/***
 * set_fqdn
 * Helper function to set the fully qualified domain name and the related
 * variables node and domain.
 */
void set_fqdn(const char *str)
{
#define qfree(v) if (v!=NULL) {\
	free(v);\
	v=NULL;\
}

	char *s, *t;

	qfree(fqdn);
	qfree(node);
	qfree(domain);

	if (strchr(str,'.')!=NULL) {
		s=strdup(str);
		fqdn=strdup(str);

		t=strchr(s,'.');
		*t='\0';
		domain=strdup(t+1);
		node=strdup(s);
		free(s);
	} else {
		fqdn=strdup(str);
		node=strdup(str);
	}
}

/***
 * xgetline
 * A hacked fgets to ignore TELNET escape sequences. Part of this
 * code was taken from wu_ftp.
 */
char *xgetline(char *s, int size, FILE *stream)
{
	register int c;
	register char *cs;

	cs=s;
retry:
	while ((c=fgetc(stream))!=EOF) {
		c&=0377;
		if (c==IAC) {
			if ((c=getc(stream))!=EOF) {
				c&=0377;
				switch (c) {
					case WILL:
					case WONT:
						c=getc(stream);
						printf("%c%c%c",IAC,DONT,0377&c);
						(void) fflush(stdout);
						continue;
					case DO:
					case DONT:
						c=getc(stream);
						printf("%c%c%c",IAC,WONT,0377&c);
						fflush(stdout);
						continue;
					case IAC:
						break;
					default:
						continue;
				}
			}
		}
		if (c=='\r')
			continue;
		*cs++=c;
		if ((--size<=0) || (c=='\n'))
			break;
	}

	if ((c==EOF) && (cs==s)) {
		if (ferror(stream) && (errno==EINTR))
			goto retry;
		return NULL;
	}

	*cs++ = '\0';

	return s;
}

/***
 * stralloc
 * Allocate size+1 bytes with calloc() and return a pointer to it.
 */
char *stralloc(int size)
{
	char *s;

	if ((s=calloc(size+1,1))==NULL) {
		perror("calloc");
		exit(255);
	}	

	return s;
}

char *wd=NULL;

/***
 * fexpand
 * Render a relative pathname into an absolute pathname.
 */
char *fexpand(char *string, int maxlen)
{
	char *str;
	char *s;
	char *string1,*string2;

	if (wd==NULL)
		wd=getcwd(NULL,MAX_DIR_LEN);

	if ((string==NULL) || (strncmp(string,"/",1)==0))
		return string;

	s=strncpy(string1=stralloc(MAX_DIR_LEN),string,MAX_DIR_LEN);
	str=strncpy(string2=stralloc(MAX_DIR_LEN),wd,MAX_DIR_LEN);
	
	s=strtok(s,"/");
	while (s!=NULL) {
		if (strcmp(s,".")==0) {
			/* NOP */
		} else if (strcmp(s,"..")==0) {
			if (strcmp(str,"/")!=0) {
				s=strrchr(str,'/');
				*s='\0';
			}
		} else if (strlen(s)>0) {
			str=strncat(str,"/",MAX_DIR_LEN);
			str=strncat(str,s,MAX_DIR_LEN);
		}

		s=strtok(NULL,"/");
	}

	string=strncpy(string,str,maxlen);
	
	free(string1); free(string2);

	return string;
}

/***
 * is_file
 * Check if a given pathname is a regular file.
 */
int is_file(const char *path)
{
	struct stat sbuf;

	if ((stat(path,&sbuf)==-1) || (!S_ISREG(sbuf.st_mode))) {
		return FALSE;
	}

	return TRUE;
}

/***
 * is_dir
 * Check if a given pathname is a directory.
 */
int is_dir(const char *path)
{
	struct stat sbuf;

	if ((stat(path,&sbuf)==-1) || (!S_ISDIR(sbuf.st_mode))) {
		return FALSE;
	}

	return TRUE;
}

/***
 * file_exists
 * Check if a given file exists.
 */
int file_exists(const char *file)
{
	return is_file(file);
}

static char *_trim_buf=NULL;
static int _trim_buf_len=0;

/***
 * trim
 * Remove unwanted characters from the beginning and the end of a
 * given string. The result is copied to the location of the input
 * string.
 */
char *trim(char *str, const char *strip)
{
	char *s;
	char *t;

	if (str==NULL)
		return NULL;

	if (strspn(str,strip)==strlen(str)) {
		*str='\0';
		return str;
	}

	if ((strlen(str)+1)>_trim_buf_len) {
		while ((strlen(str)+1)>_trim_buf_len)
			_trim_buf_len+=256;
		if (_trim_buf!=NULL)
			free(_trim_buf);
		_trim_buf=malloc(_trim_buf_len);
		*_trim_buf='\0';
	}

	/* no buffer overflow risks here; _trim_buf has enough space */
	strcpy(_trim_buf,str);
	s=_trim_buf;
	while (strchr(strip,*s)!=NULL) {
		s++;
	}

	t=s+strlen(s)-1;
	while (strchr(strip,*t)!=NULL) {
		*t='\0';
		t=s+strlen(s)-1;
	}
	str=strcpy(str,s);

	return str;
}

/***
 * lower
 * Make a character lowercase. 
 */
char lower(char c)
{
	if ((c>='A') && (c<='Z'))
		c=c+('a'-'A');

	return c;
}

/***
 * upper
 * Make a character uppercase.
 */
char upper(char c)
{
	if ((c>='a') && (c<='z'))
		c=c-('a'-'A');

	return c;
}

/***
 * strtolower
 * Make a string lowercase.
 */
char *strtolower(char *str)
{
	char *p;

	p=str;
	while (*p!='\0') {
		*p=lower(*p);
		p++;
	}

	return str;
}

/***
 * strtoupper
 * Make a string uppercase.
 */
char *strtoupper(char *str)
{
	char *p;

	p=str;
	while (*p!='\0') {
		*p=upper(*p);
		p++;
	}

	return str;
}

/***
 * capsfirst
 * Capitalize the first character of a string, make the rest
 * lowercase.
 */
char *capsfirst(char *str)
{
	str=strtolower(str);

	if (*str!='\0')
		*str=upper(*str);

	return str;
}

/***
 * background
 * Run a function with the signature int(void *) in the background,
 * and return it's return value. Optionally, any privilleges are being
 * dropped.
 */
int background(int (*func)(void *), void *opaque, int drop_priv)
{
	int pid, status;

	switch (pid=fork()) {
		case -1:
			perror("fork");
			return -1;

		case 0: /* bussiness as usual :) */
			break;

		default:
			waitpid(pid,&status,0);
			if (WIFEXITED(status))
				return WEXITSTATUS(status);
			else
				return -2;
			break;
	}

	setsid();
	chdir("/");

	if (drop_priv) {
		setgid(gid);
		setuid(uid);
	}

	exit(func(opaque));
}

#ifndef HAVE_GETHOSTNAME
# ifdef HAVE_UNAME
#  include <sys/utsname.h>
# endif /* HAVE_UNAME */
/***
 * gethostname
 * Get the system's name.
 */
int gethostname(char *name, size_t len)
{
# ifdef HAVE_UNAME
	struct utsname *buf;

	buf=malloc(sizeof(struct utsname));
	if (buf==NULL)
		return -1;
	if (uname(buf)<0)
		return -1;
	if (strlen(buf->nodename)>len) {
		errno=EINVAL;
		return -1;
	}
	strcpy(name,buf->nodename);
	free(buf);
# else /* HAVE_UNAME */
	strncpy(name,"localhost",len); /* or hard-code the real hostname */
# endif
	return 0;
}
#endif /* HAVE_GETHOSTNAME */

/* end lib.c */
