/*
	log.h -- logging routines for OMTA
	(C) Wouter Coene, 1999
	Based on log.c (C) Sam Roberts, 1999

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, write to the Free Software
	Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/*
	logging level   description
	=============== =====================================================
	OLOG_FUNC	used for log_enter() and log_leave(), and perhaps
			generic context tracking.
	OLOG_DEBUG      anything you wouldn't normally want to see in your
			syslog.
	OLOG_DIALOG     the complete SMTP dialog, usefull for somewhat more
			simple debugging.
	OLOG_STATUS     messages like 'queued message D38O547'. Short,
			simple status messages.
	OLOG_ERROR      anything that went (seriously) wrong, such as lookup
			failures.
*/

#ifndef LOG_H_
#define LOG_H_ 1

#include <stdarg.h>

/* different logging levels and -classes */

#define OLOG_NONE		0
#define OLOG_STATUS		1
#define OLOG_DIALOG		2
#define OLOG_ERROR		4
#define OLOG_DEBUG		8
#define OLOG_FUNC		16

/* This is ugly... */

#ifdef __GNUC__
# define log_status(fmt,args...)	log(OLOG_STATUS,fmt , ## args)
# define log_dialog(fmt,args...)	log(OLOG_DIALOG,fmt , ## args)
# define log_error(fmt,args...)		log(OLOG_ERROR,fmt , ## args)
# ifndef NDEBUG
#  define log_debug(fmt,args...)	log(OLOG_DEBUG,fmt , ## args)
# else /* NDEBUG */
#  define log_debug(fmt,args...)
# endif /* NDEBUG */
#else /* __GNUC__ */
# define log_alias(alias,level)		void alias(const char *fmt,...) {\
	va_list arglist;\
	va_start(arglist,fmt);\
	vlog(level,fmt,arglist);\
	va_end(arglist);\
}

# define log_alias_def(alias,level)	void alias(const char *fmt,...);

log_alias_def(log_status,OLOG_STATUS)
log_alias_def(log_dialog,OLOG_DIALOG)
log_alias_def(log_error,OLOG_ERROR)
log_alias_def(log_debug,OLOG_DEBUG)

#endif /* __GNUC__ */

#ifdef NDEBUG
# define log_enter()
# define log_leave()
#else
/*
 * the following definition was stolen from glibc2.1's assert.h, which is
 * Copyright (C) 1991,92,94,95,96,97,98,99 Free Software Foundation, Inc.
 */
# ifdef __GNUC__
#  if __GNUC__ > 2 || (__GNUC__ == 2 \
		       && __GNUC_MINOR__ >= (defined __cplusplus ? 6 : 4))
#   define __THE_FUNCTION	__PRETTY_FUNCTION__
#  else
#   define __THE_FUNCTION	((__const char *) 0)
#  endif
# else
#  if defined __STDC_VERSION__ && __STDC_VERSION__ >= 199901L
#   define __THE_FUNCTION	__func__
#  else
#   define __THE_FUNCTION	((__const char *) 0)
#  endif
# endif
# define log_enter()			log(OLOG_FUNC,"%s entered\n",__THE_FUNCTION)
# define log_leave()			log(OLOG_FUNC,"%s left\n",__THE_FUNCTION)
#endif

#ifndef NDEBUG
# define OLOG_DEFAULT	(OLOG_DEBUG|OLOG_DIALOG|OLOG_STATUS|OLOG_ERROR|OLOG_FUNC)
# define OLOG_STDERR	OLOG_DEFAULT
#else /* NDEBUG */
# define OLOG_DEFAULT	(OLOG_STATUS|OLOG_ERROR)
# define OLOG_STDERR	OLOG_ERROR
#endif /* NDEBUG */

/*
	Using perror is a Bad Thing(tm) .. but I hate to have to rewrite
	half the code ;)
*/
#define perror(x) log_error("%s: %s\n",x,strerror(errno))

extern int log_level;
extern int log_to_stderr;

void vlog(int level, const char *fmt, va_list arglist);
void log(int level, const char *fmt, ...);

#endif

/* end log.h */
