/*
	mailer.c -- main OMTA sending and queueing routines
	(C) Wouter Coene, 1999

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, write to the Free Software
	Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define WANT_SYS_TYPES
#include "config-wrap.h"
#include <sys/stat.h>

#include "mailer.h"
#include "message.h"
#include "lib.h"
#include "readconf.h"
#include "smtp.h"
#include "parser.h"
#include "log.h"
#include "execute.h"

static int mailer_add_basic_headers(t_message *message)
{
	char *buf;
	t_address *address;

	if ((message->from==NULL) || (message->to==NULL)) {
		return FALSE;
	}

	message->parse=FALSE;

	buf=stralloc(1024);

	snprintf(buf,1024,"From: %s",message->from->addr);
	message_add_body(message,buf,TRUE);

	address=message->to;
	while (address!=NULL) {
		if (address->prev==NULL)
			snprintf(buf,1024,"To: %s",address->addr);
		else
			snprintf(buf,1024,"\t%s",address->addr);

		if (address->next!=NULL)
			strncat(buf,",",1024);

		message_add_body(message,buf,TRUE);
		address=address->next;
	}

	free(buf);

	return TRUE;
}

t_message *mailer_get_message(t_message *message, FILE *fd, int do_parse, int do_period)
{
	char *buf, *s;
	int in_header;

	if (message==NULL)
		message=message_alloc();
	message->parse=do_parse;

	in_header=TRUE;
	buf=stralloc(MAX_LINE_LEN);
	forever {
		s=xgetline(buf,MAX_LINE_LEN,fd);
		if (feof(stdin) || (s==NULL))
			break;
		s=trim(s,"\r\n");

		if (do_period && (strcmp(s,".")==0))
			break;

		if ((message->bytes==0) && (strstr(s,": ")==NULL)) {
			mailer_add_basic_headers(message);
			in_header=FALSE;
		}

		if (do_period && (*s=='.'))
			s++;

		if ((message->bytes>0) && in_header && (strlen(s)==0)) {
			in_header=FALSE;
			continue;
		}

		message_add_body(message,s,in_header);

		if ((conf.max_message_size>0) && (message->bytes>conf.max_message_size)) {
			message->error=MSGERROR_NOSPACE;
			return message;
		}
	}
	free(buf);

	return message;
}

static t_line *mailer_add_header_prepend(t_message *message, const char *hdrname, int clean)
{
	return message->body;
}

static t_line *mailer_add_header_append(t_message *message, const char *hdrname, int clean)
{
	t_line *line;

	line=message->body;
	while ((line!=NULL) && (line->is_header))
		line=line->next;
	if (line==NULL)
		return message->end_body;

	return line->prev;
}

static t_line *mailer_add_header_set(t_message *message, const char *hdrname, int clean)
{
	t_line *line;

	log_enter();

	line=message->body;
	while ((line!=NULL) && (line->is_header) && (strstr(line->str,hdrname)!=line->str))
		line=line->next;
	if ((line==NULL) || (!line->is_header)) {
		log_leave();
		return NULL;
	}

	if (!clean) {
		do {
			if (line->prev!=NULL)
				line->prev->next=line->next;
			if (line->next!=NULL)
				line->next->prev=line->prev;
			line_free(line);
			if ((line=line->next)==NULL)
				break;
		} while (strchr(" \t",*(line->str))!=NULL);
	}

	log_leave();

	return line;
}

static t_line *mailer_add_header_unique(t_message *message, const char *hdrname, int clean)
{
	t_line *line;

	log_enter();

	if ((line=mailer_add_header_set(message,hdrname,FALSE))!=NULL)
		return line;

	log_leave();

	return mailer_add_header_append(message,hdrname,FALSE);
}

static t_line *mailer_add_header_enforce(t_message *message, const char *hdrname, int clean)
{
	if (mailer_add_header_set(message,hdrname,TRUE)==NULL)
		return mailer_add_header_append(message,hdrname,FALSE);

	return NULL;
}

static t_line *mailer_add_header_remove(t_message *message, const char *hdrname, int clean)
{
	mailer_add_header_set(message,hdrname,TRUE);

	return NULL;
}

static t_mailer_headerinfo mailer_headerinfo[]={
	{ HEADER_PREPEND,	mailer_add_header_prepend,	TRUE },
	{ HEADER_APPEND,	mailer_add_header_append,	FALSE },
	{ HEADER_SET,		mailer_add_header_set,		TRUE },
	{ HEADER_UNIQUE,	mailer_add_header_unique,	FALSE },
	{ HEADER_ENFORCE,	mailer_add_header_enforce,	FALSE },
	{ HEADER_REMOVE,	mailer_add_header_remove,	FALSE },
	{ 0,			NULL,				0 }
};


static void mailer_add_headers(t_message *message)
{
	int i, t, prepend;
	char *p, *s, *h;
	t_line *start_line, *end_line;
	t_line *line, *where;

	log_enter();

	message->parse=FALSE;
	for (i=0; i<conf.extra_header_count; i++) {
		/* reset variables */
		start_line=NULL;
		end_line=NULL;

		/* parse the header to be added */
		p=parse(message,conf.extra_header[i],FALSE);

		if (strchr(p,':')==NULL) {
			free(p);
			continue;
		}

		s=strdup(p);

		/* get the header name, including ':' (need that later) */
		*(strchr(p,':')+1)='\0';
		h=strdup(p);

		free(p);

		/* see what we have to do with the header */
		t=0;
		where=NULL;
		prepend=FALSE;
		while (mailer_headerinfo[t].handler!=NULL) {
			if (mailer_headerinfo[t].type==conf.extra_header_type[i]) {
				where=mailer_headerinfo[t].handler(message,h,FALSE);
				prepend=mailer_headerinfo[t].prepend;
				break;
			}
			t++;
		}

		if (where==NULL) {
			free(s);
			continue;
		}

		/* split it up into separate t_line structs, placed in a linked list */
		p=s;
		s=strtok(p,"\n");
		while (s!=NULL) {
			/* very dirty hack.. */
			if (s!=p)
				*(--s)='\t';

			line=line_alloc(s,TRUE);

			if (start_line==NULL) {
				start_line=line;
				end_line=line;
			} else {
				end_line->next=line;
				line->prev=end_line;
				end_line=line;
			}

			s=strtok(NULL,"\n");
		}
		free(p);

		/* go add it */
		message_add_body_raw(message,start_line,where,prepend);

		/* free variables */
		free(h);
	}

	log_leave();
}

static int mailer_relay(t_message *message)
{
	FILE *fd;

	log_enter();

	if (conf.relay_host!=NULL) {
		fd=smtp_connect(conf.relay_host,conf.relay_port);
		if (fd!=NULL) {
			if (smtp_send(message,fd,NULL)) {
				smtp_close(fd);
				log_status("Relayed %s\n",message->id);
				return TRUE;
			}
			smtp_close(fd);
			log_error("Relay of message with id %s failed\n"
				"Server said: %s\n",message->id,last_message);
		}
	}

	log_leave();

	return FALSE;
}

static int mailer_deliver(t_message *message)
{
	char *s;
	FILE *fd;
	int skipped, delivered;
	int saved_pid, status;
	t_address *address;

	if ((conf.local_mta==NULL) || (conf.local_count==0))
		return FALSE;

	log_enter();

	skipped=FALSE;
	delivered=FALSE;
	address=message->to;
	while (address!=NULL) {
		if (!address_islocal(address,conf.local,conf.local_count)) {
			skipped=TRUE;
			address=address->next;
			continue;
		}
		if (address->handled) {
			address=address->next;
			continue;
		}

		address->active=TRUE;

		s=parse(message,conf.local_mta,TRUE);
		fd=pexecute(s,'w',conf.insecure_mailer,&saved_pid);
		free(s);

		if (fd==NULL) {
			log_error("Error delivering %s to local user %s: %s\n",
					message->id,address->user,strerror(errno));
			address->active=FALSE;
			address=address->next;
			continue;
		}

		message_out(message,fd);
		if ((status=pdone(fd,saved_pid))>0) {
			log_error("Error delivering %s to local user %s; LocalMTA returned %d\n",
					message->id,address->user,status);
			address->active=FALSE;
			address=address->next;
			continue;
		}

		address->active=FALSE;
		address->handled=TRUE;

		delivered=TRUE;

		address=address->next;
	}

	if (delivered)
		log_status("Delivered %s locally\n",message->id);

	log_leave();

	return !skipped;
}

static int mailer_queue(t_message *message)
{
	FILE *fd;
	char *s;
	int l, ok;

	l=strlen(conf.queue_path)+strlen(message->id)+1;

	s=stralloc(strlen(conf.queue_path)+strlen(message->id)+1);
	snprintf(s,l+1,"%s/%s",conf.queue_path,message->id);

	if ((fd=fopen(s,"w"))==NULL) {
		log_error("Couldn't write to %s\n",s);
		free(s);
		return FALSE;
	}

	ok=message_write(message,fd);
	fclose(fd);

	l=S_IRUSR|S_IWUSR;
	if (gid!=egid)
		l|=S_IRGRP|S_IWGRP;
	chmod(s,l);

	free(s);

	if (ok)
		log_status("Queued %s\n",message->id);

	return ok;
}

int mailer_send(t_message *message)
{
	mailer_add_headers(message);
	return mailer_deliver(message)
		|| mailer_relay(message)
		|| background((void *) mailer_queue,message,conf.insecure_spool);
}

/* end mailer.c */
