/*
	message.c -- generic message handling
	(C) Wouter Coene, 1999

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, write to the Free Software
	Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define WANT_TIME
#define WANT_UNISTD
#include "config-wrap.h"

#include "message.h"
#include "lib.h"
#include "omta.h"
#include "log.h"
#include "line.h"
#include "header.h"

static short int message_count=0;

/***
 * readstring
 * Read a string.
 */
static char *readstring(char *s, FILE *fd)
{
	char *t;

	*s='\0';
	t=trim(fgets(s,1024,fd),"\n\r");

	return t;
}

/***
 * message_set_id
 * Generate a new message id. The message id is guaranteed to be unique even
 * over NFS mounts.
 */
static t_message *message_set_id(t_message *message)
{
	time_t tm;
	pid_t pid;
	int l;

	if (message==NULL)
		return message;

	log_enter();

	tm=time(NULL);
	pid=getpid();

	l=strlen(node)+24;
	message->id=stralloc(l);
	snprintf(message->id,l+1,"%s.%d-%d.%d",node,(int) pid,message_count++,(int) tm);

	log_leave();

	return message;
}

/***
 * message_next_address
 * Return a pointer to the next email address in a comma separated
 * list.
 */
static char *message_next_address(char *str)
{
	char *p, *t;

	log_enter();

	p=str;
	forever {
		if (*p=='\0') {
			p=NULL;
			break;
		}

		if (*p=='"') {
			if ((t=strchr(p+1,'"'))!=NULL) {
				p=t+1;
			}
		} else if (*p=='(') {
			if ((t=strchr(p+1,')'))!=NULL) {
				p=t+1;
			}
		} else if (*p==',')
			break;

		p++;
	}

	if (p!=NULL) {
		*p='\0';
		p++;
	}

	log_leave();

	return p;
}

/***
 * message_do_parse
 * Do the actual parsing for message_parse_header. Also called by
 * message_add_body directly when headers are done.
 */
static t_message *message_do_parse(t_message *message, t_header *header)
{
	char *s, *p, *n;

	if ((message==NULL) || (header==NULL))
		return NULL;

	log_enter();

	if (strcasecmp(header->name,"from")==0) {
		if (message->from==NULL)
			message_set_sender(message,address_alloc(header->content));
	} else if (strcasecmp(header->name,"to")==0) {
		s=strdup(header->content);
		p=s;
		while (p!=NULL) {
			n=message_next_address(p);
			p=trim(p," \t");
			message_add_recipient(message,address_alloc(p));
			p=n;
		}

		free(s);
	}

	log_leave();

	return message;
}

/***
 * message_parse_header
 * Parse a header.
 */
static t_message *message_parse_header(t_message *message, const char *str)
{
	t_header *header;

	if ((message==NULL) || (str==NULL) || (!message->parse))
		return NULL;

	log_enter();

	if ((header=header_create(str,message->__last_header))!=NULL) {
		if (header!=message->__last_header) {
			if (message->__last_header!=NULL) {
				message_do_parse(message,message->__last_header);
				header_free(message->__last_header);
			}
			message->__last_header=header;
		}
	}

	log_leave();

	return message;
}

/***
 * message_alloc
 * Allocate a new t_message struct.
 */
t_message *message_alloc()
{
	return message_set_id(calloc(sizeof(t_message),1));
}

/***
 * message_free
 * Free a previously allocated t_message struct.
 */
t_message *message_free(t_message *message)
{
	t_address *address;
	t_line *line;
	void *next;

	if (message->from!=NULL)
		address_free(message->from);

	address=message->to;
	while (address!=NULL) {
		next=address->next;
		address_free(address);
		address=next;
	}

	line=message->body;
	while (line!=NULL) {
		next=line->next;
		line_free(line);
		line=next;
	}

	if (message->helo!=NULL)
		free(message->helo);
	if (message->id!=NULL)
		free(message->id);

	free(message);

	return NULL;
}

/***
 * message_set_sender
 * Set the sender address of the message.
 */
t_message *message_set_sender(t_message *message, t_address *sender)
{
	log_enter();

	if (message!=NULL) {
		if (message->from==NULL) {
			log_debug("Sender set to %s\n",sender->addr);
			message->from=sender;
		}
	}

	log_leave();

	return message;
}

/***
 * message_add_recipient
 * Add a recipient to a message.
 */
t_message *message_add_recipient(t_message *message, t_address *recipient)
{
	if ((message==NULL) || (recipient==NULL))
		return NULL;

	log_debug("Added recipient %s\n",recipient->addr);

	if (message->end_to==NULL) {
		message->to=recipient;
		message->end_to=recipient;
	} else {
		message->end_to->next=recipient;
		recipient->prev=message->end_to;
		message->end_to=recipient;
	}

	return message;
}

/***
 * message_add_recipients
 * Add a list of recipients to a message.
 */
t_message *message_add_recipients(t_message *message, t_address *recipient)
{
	t_address *next;

	if (message==NULL)
		return NULL;

	while (recipient!=NULL) {
		next=recipient->next;
		message_add_recipient(message,recipient);
		recipient=next;
	}

	return message;
}

/***
 * message_remove_recipients
 * Remove recipients from a message, based on a linked list of addresses.
 */
t_message *message_remove_recipients(t_message *message, t_address *recipient)
{
	t_address *addr1, *addr2;

	if ((message==NULL) || (recipient==NULL))
		return NULL;

	log_enter();

	addr1=message->to;
	while (addr1!=NULL) {
		addr2=recipient;
		while (addr2!=NULL) {
			if (addr1->formatted==NULL)
				address_format(addr1);
			if (addr2->formatted==NULL)
				address_format(addr2);
			if (strcasecmp(addr1->formatted,addr2->formatted)==0) {
				log_debug("Removed recipient %s\n",addr1->addr);
				addr1->handled=TRUE;
			}
			addr2=addr2->next;
		}
		addr1=addr1->next;
	}

	log_leave();

	return message;
}

/***
 * message_set_helo
 * Set the HELO field of a message.
 */
t_message *message_set_helo(t_message *message, const char *str)
{
	log_enter();

	if (message!=NULL) {
		if (message->helo!=NULL)
			free(message->helo);
		message->helo=strdup(str);
	}

	log_leave();

	return message;
}

/***
 * message_add_body_raw
 * Add a line, or a set of lines, to the message body (== headers + text).
 */
t_message *message_add_body_raw(t_message *message, t_line *line, t_line *where, int prepend)
{
	t_line *last;

	if ((message==NULL) || (line==NULL) || ((message->body!=NULL) && (where==NULL)))
		return NULL;

	last=line;
	while (last->next!=NULL)
		last=last->next;

	if (message->body==NULL) {
		message->body=line;
		message->end_body=last;
	} else {
		if (prepend) {
			line->prev=where->prev;
			where->prev=last;
			last->next=where;
			if (line->prev!=NULL)
				line->prev->next=line;
			else
				message->body=line;
		} else {
			last->next=where->next;
			where->next=line;
			line->prev=where;
			if (last->next!=NULL)
				last->next->prev=last;
			else
				message->end_body=last;
		}
	}

	return message;
}

/***
 * message_add_body
 * Add a line to the message body (== headers + text).
 */
t_message *message_add_body(t_message *message, const char *str, int is_header)
{
	if ((message!=NULL) && (str!=NULL)) {
		if (is_header && message->parse) {
			message_parse_header(message,str);
		} else if ((message->parse) && (message->end_body!=NULL)
				&& (is_header!=message->end_body->is_header)
				&& (message->__last_header!=NULL)) {
			message_do_parse(message,message->__last_header);
			header_free(message->__last_header);
		}

		message->bytes+=strlen(str);
		if (!is_header)
			message->body_bytes+=strlen(str);

		message_add_body_raw(message,line_alloc(str,is_header),message->end_body,FALSE);

		message->lines++;
		if (!is_header)
			message->body_lines++;
	}

	return message;
}

/***
 * message_out
 * Write the body of the t_message struct to the given filedescriptor.
 */
int message_out(t_message *message, FILE *fd)
{
	int last_was_header;
	t_line *line;

	if ((message==NULL) || (fd==NULL))
		return FALSE;

	last_was_header=TRUE;
	line=message->body;
	while (line!=NULL) {
		if (line->is_header!=last_was_header) {
			last_was_header=FALSE;
			fprintf(fd,"\n");
		}
		fprintf(fd,"%s\n",line->str);
		line=line->next;
	}

	return TRUE;
}

/***
 * message_write
 * Write the t_message struct to the given filedescriptor.
 */
int message_write(t_message *message, FILE *fd)
{
	t_address *address;

	if (message==NULL) {
		log_debug("message_write called with NULL t_message\n");
		return FALSE;
	}
	if ((message->from==NULL) || (message->to==NULL)) {
		log_error("Cannot queue %s; no sender and/or recipients\n",message->id);
		return FALSE;
	}

	fprintf(fd,MESSAGE_SIGNATURE "%d\n%s\n",MESSAGE_VERSION,message->id);

	if (message->from->formatted==NULL)
		address_format(message->from);
	fprintf(fd,"%s\n",message->from->formatted);

	address=message->to;
	while (address!=NULL) {
		if (!address->handled) {
			if (address->formatted==NULL)
				address_format(address);
			fprintf(fd,"%s\n",address->formatted);
		}
		address=address->next;
	}
	fprintf(fd,"-\n");

	return message_out(message,fd);
}

/***
 * message_read
 */
t_message *message_read(FILE *fd, int do_body)
{
	t_message *message;
	char *buf, *s;
	int ver, in_header;

	message=message_alloc();
	buf=stralloc(1024);

	message->parse=FALSE;

	readstring(buf,fd);
	if (strstr(buf,MESSAGE_SIGNATURE)==buf) {
		ver=atoi(buf+strlen(MESSAGE_SIGNATURE));
		if (ver>MESSAGE_VERSION) {
			log_error("Cannot handle messages of version %d\n",ver);
			return NULL;
		}

		free(message->id);
		message->id=strdup(readstring(buf,fd));

		message_set_sender(message,address_alloc(readstring(buf,fd)));
		forever {
			readstring(buf,fd);
			if (strcmp(buf,"-")==0)
				break;
			message_add_recipient(message,address_alloc(buf));
		}
	} else if (strlen(buf)==0) {
		log_error("Empty or damaged message file detected; please repair\n");
		return NULL;
	} else {
		message_set_sender(message,address_alloc(buf));

		s=strtok(readstring(buf,fd),MESSAGE_OLDDELIM);
		while (s!=NULL) {
			message_add_recipient(message,address_alloc(s));
			s=strtok(NULL,MESSAGE_OLDDELIM);
		}		
	}

	if (do_body) {
		in_header=TRUE;
		while (readstring(buf,fd)!=NULL) {
			if ((!in_header) && (strlen(buf)==0))
				in_header=FALSE;
			else
				message_add_body(message,buf,in_header);
		}
	}

	free(buf);

	return message;
}

/* end message.c */
