/*
	parser.c -- string parser for OMTA
	(C) Wouter Coene, 1999

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, write to the Free Software
	Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define WANT_UNISTD
#define WANT_TIME
#include "config-wrap.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
                     
#include "parser.h"
#include "omta.h"
#include "message.h"
#include "lib.h"

/***
 * stradd
 * Append a string to another string.
 */
static void stradd(char *s, const char *add)
{
	int l;

	if (add==NULL)
		return;

	l=strlen(add);
	if (l+strlen(s)>MAX_LINE_LEN)
		l=strlen(s)-MAX_LINE_LEN;

	strncat(s,add,l);
}

/***
 * strcadd
 * Append a character to a string.
 */
static void strcadd(char *s, char add)
{
	if (strlen(s)!=MAX_LINE_LEN)
		strncat(s,&add,1);
}

static struct tm *now;
static time_t tnow;

/***
 * date
 * Get the date.
 */
static void date()
{
	tnow=time(NULL);
	now=localtime(&tnow);
}

/***
 * tzoffset
 * Calculate the timezone offzet (lifted directly from fetchmail).
 */
static char *tzoffset(time_t *timenow)
{
	static char offset_string[6];
	struct tm gmt, *lt;
	int off;
	char sign = '+';

	gmt = *gmtime(timenow);
	lt = localtime(timenow);
	off = (lt->tm_hour - gmt.tm_hour) * 60 + lt->tm_min - gmt.tm_min;
	if (lt->tm_year < gmt.tm_year)
		off -= 24 * 60;
	else if (lt->tm_year > gmt.tm_year)
		off += 24 * 60;
	else if (lt->tm_yday < gmt.tm_yday)
		off -= 24 * 60;
	else if (lt->tm_yday > gmt.tm_yday)
		off += 24 * 60;
	if (off < 0) {
		sign = '-';
		off = -off;
	}
	if (off >= 24 * 60)			/* should be impossible */
	   off = 23 * 60 + 59;		/* if not, insert silly value */
	sprintf(offset_string, "%c%02d%02d", sign, off / 60, off % 60);

	return (offset_string);
} 

/***
 * conditional
 * Process a conditional expression.
 */
static char *conditional(t_message *message, char *p, char *s, int machine_readable)
{
	char n, *b, *t;
	int i, test;

	p++;
	n=*p;
	if ((n=='\0') || (strchr("rsdRqom",n)==NULL))
		return p-1;

	p++;
	if (*p!='{')
		return p-1;

	i=0;
	while ((*p!='}') && (*p!='\0')) {
		i++; p++;
	}

	b=stralloc(i);
	snprintf(b,i,"%s",(p-i)+1);

	test=FALSE;
	switch (n) {
		case 'r':
			test=message->from_host.s_addr!=0;
			break;

		case 's':
			test=message->from_host.s_addr==0;
			break;

		case 'o':
			test=address_count(message->to)==1;
			break;

		case 'm':
			test=address_count(message->to)>1;
			break;

		default:
			test=FALSE;
	}

	if (test) {
		t=parse(message,b,machine_readable);
		stradd(s,t);
		free(t);
	}
	free(b);

	if (*p=='\0')
		p--;

	return p;
}

/***
 * parse
 * Parse a printf() like string. See the parser documentation for
 * more details.
 */
char *parse(t_message *message, const char *fmt, int machine_readable)
{
	int i, need_active, get_username, get_domainname;
	char *f, *p, *s, *t, *b;
	struct hostent *h;
	t_address *address;

	s=stralloc(MAX_LINE_LEN);
	f=strdup(fmt);
	p=f;
	while (*p!='\0') {
		if (*p=='%') {
			need_active=FALSE;
			get_username=FALSE;
			get_domainname=FALSE;

			switch (*++p) {
				case 'i':
					if (message->id!=NULL)
						stradd(s,message->id);
					else
						stradd(s,"(n/a)");
					break;

				case 'n':
					stradd(s,node);
					break;

				case 'd':
					stradd(s,domain);
					break;

				case 'f':
					stradd(s,fqdn);
					break;

				case 'N':
					date();
					t=stralloc(40);
					strftime(t,40,TIME_FORMAT,now);
					strncpy(strstr(t,"XXXXX"),tzoffset(&tnow),5);
					stradd(s,t);
					free(t);
					break;

				case 'D':
					date();
					t=stralloc(32);
					strftime(t,32,TIME_DATEONLY,now);
					strncpy(strstr(t,"XXXXX"),tzoffset(&tnow),5);
					stradd(s,t);
					free(t);
					break;

				case 'T':
					date();
					t=stralloc(32);
					strftime(t,32,TIME_TIMEONLY,now);
					strncpy(strstr(t,"XXXXX"),tzoffset(&tnow),5);
					stradd(s,t);
					free(t);
					break;

				case 'S':
					p++;
					if (*p!='{') {
						p--;
						break;
					}

					i=0;
					while ((*p!='}') && (*p!='\0')) {
						i++; p++;
					}

					b=stralloc(i);
					snprintf(b,i,"%s",(p-i)+1);

					date();
					t=stralloc(64);
					strftime(t,64,b,now);
					stradd(s,t);
					free(t);
					free(b);
					break;

				case 's':
					if (message->from->formatted==NULL)
						address_format(message->from);
					stradd(s,message->from->formatted);
					break;

				case 'u':
					stradd(s,user);
					break;

				case 'g':
					stradd(s,user_gecos);
					break;

				case 'F':
					if (message->from_host.s_addr>0) {
						stradd(s,inet_ntoa(message->from_host));
					}
					break;

				case 'h':
					if (message->from_host.s_addr==0) {
						stradd(s,machine_readable? "stdin":"standard input");
					} else {
						h=gethostbyaddr((char *) &message->from_host,
							sizeof(struct in_addr),AF_INET);
						if (h!=NULL) {
							stradd(s,h->h_name);
						} else {
							stradd(s,inet_ntoa(message->from_host));
						}
					}
					break;

				case '!':
					get_username=TRUE;
				case '@':
					if (!get_username)
						get_domainname=TRUE;
				case 't':
					need_active=TRUE;
				case 'r':
					address=message->to;
					i=0;
					while (address!=NULL) {
						if (need_active && (!address->active)) {
							address=address->next;
							continue;
						}
						if (address->formatted==NULL)
							address=address_format(address);
						if (i>0) {
							if ((address->next==NULL)
									&& (!machine_readable))
								stradd(s," and ");
							else
								stradd(s,machine_readable? ",":", ");
						}
						if (get_username) {
							stradd(s,address->user);
						} else if (get_domainname) {
							stradd(s,address->domain);
						} else {
							stradd(s,address->formatted);
						}

						address=address->next;
						i++;
					}
					break;

				case 'H':
					stradd(s,message->helo);
					break;

				case 'o':
					stradd(s,NAME);
					break;

				case 'v':
					stradd(s,VERSION);
					break;

				case 'V':
					stradd(s,SHORTVERSION);
					break;

				case 'U':
					stradd(s,OMTA_URI);
					break;

				case '?':
					p=conditional(message,p,s,machine_readable);
					break;

				case 'c':
					t=stralloc(10);
					snprintf(t,11,"%d",message->bytes);
					stradd(s,t);
					free(t);
					break;

				case 'C':
					t=stralloc(10);
					snprintf(t,11,"%d",message->body_bytes);
					stradd(s,t);
					free(t);
					break;

				case 'l':
					t=stralloc(10);
					snprintf(t,11,"%d",message->lines);
					stradd(s,t);
					free(t);
					break;

				case 'L':
					t=stralloc(10);
					snprintf(t,11,"%d",message->body_lines);
					stradd(s,t);
					free(t);
					break;

				case 'b':
					strcadd(s,'\n');
					break;

				case '%':
					strcadd(s,'%');
					break;

				default:
					break;
			}
		} else {
			strcadd(s,*p);
		}

		p++;
	}

	free(f);

	t=strdup(s);
	free(s);

	return t;
}

/* parser.c */
