/*
	queue.c -- queue processing routines (dump and transfer)
	(C) Wouter Coene, 1999

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, write to the Free Software
	Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define WANT_UNISTD
#define WANT_DIRENT
#include "config-wrap.h"

#include "queue.h"
#include "message.h"
#include "address.h"
#include "log.h"
#include "omta.h"
#include "smtp.h"
#include "readconf.h"
#include "lib.h"

static int sent=0;

/***
 * mail_select
 * Selection routine for scandir().
 */
static int mail_select(SCANDIR_CONST struct dirent *d)
{
	return (*(d->d_name)!='.') && (*(d->d_name+strlen(d->d_name)-1)!='~');
}

/***
 * process_queue
 * Do something with the queue.
 */
static int process_queue(const char *dir, t_queue_file_handler handler, void *opaque, int need_write)
{
	int n, total, ok, mask;
	struct dirent **namelist;
	char *filename;
	FILE *fd;

	log_enter();

	mask=R_OK;
	if (need_write)
		mask|=W_OK;
	if (!is_dir(dir)) {
		log_error("%s: not found or not a directory\n",dir);
		log_leave();
		return 0;
	}
	if (conf.insecure_spool)
		if (access(dir,mask)<0) {
			log_error("%s: access denied\n",dir);
			log_leave();
			return 0;
		}

	total=0;
	ok=TRUE;

	filename=stralloc(1024);
	n=scandir(dir,&namelist,mail_select,NULL);
	while (n>0) {
		if (handler!=NULL) {
			snprintf(filename,1024,"%s/%s",dir,namelist[n-1]->d_name);

			if ((conf.insecure_spool && (access(filename,mask)<0))
					|| ((fd=fopen(filename,"r"))==NULL)) {
				log_error("Cannot open %s\n",filename);
				ok=FALSE;
				n--;
				continue;
			}

			ok=handler(fd,filename,n,opaque);
			fclose(fd);
			if (!ok)
				break;
		}
		n--; total++;
	}

	if ((total==0) && ok)
		printf("Queue is empty\n");

	free(filename);

	log_leave();

	return total;
}

/***
 * dump_handler
 * Dump out a given message.
 */
static int dump_handler(FILE *fd, const char *filename, int n, void *opaque)
{
	t_message *message;
	t_address *address;
	int t;

	if ((message=message_read(fd,FALSE))==NULL) {
		log_error("Reading of message file %s failed\n",filename);
		return FALSE;
	}

	printf("Message id %s\n",message->id);

	if (message->from->formatted==NULL)
		message->from=address_format(message->from);
	printf("\tFrom:\t%s\n",message->from->formatted);

	printf("\tTo:\t");
	t=16;
	address=message->to;
	while (address!=NULL) {
		if (address->formatted==NULL)
			address=address_format(address);

		t+=strlen(address->formatted);
		if (t>73) {
			printf("%s\n\t\t",address!=message->to? ",":"");
			t=16+strlen(address->formatted);
		} else if (address!=message->to) {
			printf(", ");
		}
		printf("%s",address->formatted);

		address=address->next;
	}
	printf("\n");
	if (n>1)
		printf("\n");

	message_free(message);

	return TRUE;
}

/***
 * mode_dump
 * Handler routine for queue dump mode.
 */
int mode_dump()
{
	return process_queue(conf.queue_path,dump_handler,NULL,FALSE)? TRUE:FALSE;
}

static int bytes_done, periods_done;

/***
 * xfer_status
 * Print dots.
 */
static void xfer_status(t_line *line)
{
	bytes_done+=strlen(line->str);
	if (bytes_done>(periods_done*1024)) {
		periods_done++;
		printf(".");
	}
}

/***
 * xfer_handler
 * Transfer a given message.
 */
static int xfer_handler(FILE *fdin, const char *filename, int n, void *fdout)
{
	t_message *message;

	message=message_read(fdin,TRUE);

	printf("Sending message %d (%d bytes) ",sent+1,message->bytes);
	bytes_done=0;
	periods_done=0;
	if (smtp_send(message,(FILE *) fdout,xfer_status)) {
		sent++;
		log_status("Message with id %s sent\n",message->id);
		unlink(filename);
		printf(" done\n");
	} else {
		printf(" error, server said: %s\n",last_message);
		log_status("Error sending message with id %s; "
			"The server said: %s\n",message->id,last_message);
	}

	message_free(message);

	return TRUE;
}

/***
 * mode_xfer
 * Handler routine for queue transfer mode.
 */
int mode_xfer()
{
	FILE *fd;
	int total;

	if (process_queue(conf.queue_path,NULL,NULL,TRUE)==0)
		return FALSE;

	if ((fd=smtp_connect(conf.smtp_server,conf.smtp_port))==NULL) {
		log_error("Unable to connect to server %s port %d\n",conf.smtp_server,conf.smtp_port);
		return FALSE;
	}
	printf("Connected to server %s port %d\n",conf.smtp_server,conf.smtp_port);

	total=process_queue(conf.queue_path,xfer_handler,fd,TRUE);
	printf("%d of %d messages sent\n",sent,total);

	smtp_close(fd);
	printf("Connection closed\n");

	return TRUE;
}

/* end queue.c */
