/*
	readconf.c -- configuration reading routines
	(C) Wouter Coene, 1999

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, write to the Free Software
	Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define WANT_UNISTD
#include "config-wrap.h"
#include <pwd.h>
#include <sys/types.h>

#define CONFIG_HAVE_DATA
#include "readconf.h"
#include "omta.h"
#include "lib.h"
#include "log.h"

char *conf_global=NULL;
char *conf_local=NULL;
t_config conf;

/***
 * config_distate
 * Decode values like 'yes' and 'true'.
 */
static int config_distate(const char *str)
{
	if ((strcasecmp(str,"yes")==0)
			|| (strcasecmp(str,"true")==0)
			|| (strcasecmp(str,"on")==0)
			|| (strcasecmp(str,"1")==0)) {

		return TRUE;
	} else {
		return FALSE;
	}
}

/***
 * config_func
 * Handler function for getconf.
 */
static int config_func GETCONF_CONFIG_PARAM
{
#define check_args(m)	if (argc!=(m+1)) {\
	log_error("Invalid no. of arguments for setting %s in %s at line %d\n"\
		"expected %d, got %d\n",argv[0],file,lineno,m,argc);\
	ok=FALSE;\
	break;\
}
#define check_args_minmax(min,max) if ((argc<(min+1)) || (argc>(max+1))) {\
	log_error("Invalid no. of arguments for setting %s in %s at line %d\n"\
		"expected at least %d and at most %d, got %d\n",\
		argv[0],file,lineno,min,max,argc);\
	ok=FALSE;\
	break;\
}
#define check_hasargs() if (argc<2) {\
	log_error("Arguments ommitted for setting %s in %s at line %d\n",argv[0],file,lineno);\
	ok=FALSE;\
	break;\
}

	int ok, i;
	char *s;

	ok=TRUE;
	switch (number) {
		case CONF_HOSTNAME:
			check_args(1);

			if (conf.host_name!=NULL)
				free(conf.host_name);
			conf.host_name=strdup(argv[1]);
			break;

		case CONF_QUEUEDIR:
			check_args(1);

			if (conf.is_local)
				conf.insecure_spool=TRUE;

			if (conf.queue_path!=NULL)
				free(conf.queue_path);

			s=stralloc(MAX_DIR_LEN);
			strncpy(s,argv[1],MAX_DIR_LEN);
			fexpand(s,MAX_DIR_LEN);

			conf.queue_path=strdup(s);
			free(s);

			if (conf.is_local)
				conf.insecure_spool=TRUE;
			break;

		case CONF_SMTPSERVER:
			check_args_minmax(1,2);

			if (conf.smtp_server!=NULL)
				free(conf.smtp_server);
			conf.smtp_server=strdup(argv[1]);
			if (argc>2)
				conf.smtp_port=atoi(argv[2]);
			break;

		case CONF_RELAYTO:
			check_args_minmax(1,2);

			if (conf.relay_host!=NULL)
				free(conf.relay_host);
			conf.relay_host=strdup(argv[1]);
			if (argc>2)
				conf.relay_port=atoi(argv[2]);
			break;

		case CONF_LOCAL:
			check_hasargs();

			for (i=0; i<(argc-1); i++) {
				conf.local_count++;
				conf.local=realloc(conf.local,sizeof(void *)*conf.local_count);
				conf.local[conf.local_count-1]=strdup(argv[i+1]);
			}
			break;

		case CONF_LOCALMTA:
			check_args(1);

			if (conf.is_local)
				conf.insecure_mailer=TRUE;

			if (conf.local_mta!=NULL)
				free(conf.local_mta);
			conf.local_mta=strdup(argv[1]);

			if (conf.is_local)
				conf.insecure_mailer=TRUE;
			break;

		case CONF_RELAYFOR:
			check_hasargs();

			for (i=0; i<(argc-1); i++) {
				conf.relay_for_count++;
				conf.relay_for=realloc(conf.relay_for,sizeof(void *)*conf.relay_for_count);
				conf.relay_for[conf.relay_for_count-1]=strdup(argv[i+1]);
			}
			break;

		case CONF_SMTPGREET:
			check_args(1);

			if (conf.smtp_greeting!=NULL)
				free(conf.smtp_greeting);
			conf.smtp_greeting=strdup(argv[1]);
			break;

		case CONF_SERVERPORT:
			check_args(1);

			if (conf.is_local)
				conf.insecure_port=TRUE;

			conf.server_port=atoi(argv[1]);
			if (conf.is_local)
				conf.insecure_port=TRUE;
			break;

		case CONF_INTERFACE:
			check_args(1);

			if (conf.interface!=NULL)
				free(conf.interface);
			conf.interface=strdup(argv[1]);
			break;

		case CONF_ALLOWLOCAL:
			check_args(1);

			conf.allow_local=config_distate(argv[1]);
			break;

		case CONF_FROMADDRESS:
			check_args(1);

			/* stop evil sysadmins :^) */
			if (!conf.is_local) {
				log_error("Setting %s cannot be set in global configuration file\n",argv[0]);
				ok=FALSE;
				break;
			}

			if (conf.from_address!=NULL)
				free(conf.from_address);
			conf.from_address=strdup(argv[1]);
			break;

		case CONF_MAXMSGSIZE:
			check_args(1);

			if (conf.is_local) {
				log_error("Setting %s cannot be set in local configuration file\n",argv[0]);
				ok=FALSE;
				break;
			}

			conf.max_message_size=atoi(argv[1]);
			break;

		case CONF_LOGGING:
			check_args(1);

			conf.logging=config_distate(argv[1]);
			break;

		case CONF_LOGLEVEL:
			check_args(1);

			log_level=atoi(argv[1]);
			break;

		case CONF_PREPENDHEADER:
		case CONF_APPENDHEADER:
		case CONF_SETHEADER:
		case CONF_UNIQUEHEADER:
		case CONF_ENFORCEHEADER:
		case CONF_REMOVEHEADER:
			check_args(1);

			conf.extra_header_count++;

			conf.extra_header=realloc(conf.extra_header,sizeof(void *)*conf.extra_header_count);
			conf.extra_header[conf.extra_header_count-1]=strdup(argv[1]);

			conf.extra_header_type=realloc(conf.extra_header_type,sizeof(int)*conf.extra_header_count);
			conf.extra_header_type[conf.extra_header_count-1]=number;
			break;

		case -1:
			log_error("Invalid setting %s in %s at line %d\n",argv[0],file,lineno);
			ok=FALSE;
			break;

		default:
			log_error("readconf.c broken; got number %d\n",number);
			ok=FALSE;
			break;
	}

	return ok;
}

/***
 * config_read
 * Read the given configuration file.
 */
static int config_read(const char *file, int local)
{
	int ok;

	log_enter();

	if (local && (access(file,R_OK)<0)) {
		log_error("%s: Access denied\n",file);
		return FALSE;
	}

	conf.is_local=local;
	ok=getconf(getconf_data,file,config_func)>0;

	log_leave();

	return ok;
}

/***
 * config
 * Read the global and local configuration files (if any).
 */
int config(int no_local)
{
	int i, is_local, ok;

	ok=TRUE;

	log_enter();
	if (conf_global==NULL) {
		is_local=FALSE;
		conf_global=strdup(CONFIG_GLOBAL);
	} else {
		is_local=strcmp(CONFIG_GLOBAL,conf_global)==0;
	}

	if ((!no_local) && (conf_local==NULL)) {
		i=strlen(user_home)+strlen(CONFIG_LOCAL)+1;
		conf_local=stralloc(i);
		strcat(conf_local,user_home);
		strcat(conf_local,"/");
		strcat(conf_local,CONFIG_LOCAL);
	}

	memset(&conf,'\0',sizeof(t_config));

	/* set defaults */
	conf.queue_path=strdup(CONF_QUEUE);
	conf.smtp_server=strdup(CONFIG_DEF_SERVER);
	conf.smtp_port=CONFIG_DEF_SMTPPORT;
	conf.relay_port=CONFIG_DEF_SMTPPORT;
	conf.server_port=CONFIG_DEF_SMTPPORT;
	conf.smtp_greeting=strdup(CONFIG_DEF_GREETING);
	conf.max_message_size=CONFIG_DEF_MAXMSGSIZE;
	conf.logging=CONFIG_DEF_LOGGING;

	ok=config_read(conf_global,is_local);
	if ((!no_local) && ok && (conf_local!=NULL) && (!is_local)
			&& (conf.allow_local) && file_exists(conf_local))
		ok=config_read(conf_local,TRUE);

	if (!ok) {
		log_error("Error while reading configuration file\n");
	}

	log_leave();

	return ok;
}

/* end readconf.c */
