/*
	smtp.c -- SMTP routines for OMTA
	(C) Wouter Coene, 1999

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, write to the Free Software
	Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define WANT_UNISTD
#define WANT_SYS_TYPES
#include "config-wrap.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "smtp.h"
#include "mailer.h"
#include "omta.h"
#include "header.h"
#include "address.h"
#include "lib.h"
#include "log.h"
#include "readconf.h"
#include "canon-host.h"

char *last_message=NULL;
char *buf=NULL;

/***
 * smtp_expect
 * Read a response from the SMTP server and check if it's the
 * response we're expecting.
 */
int smtp_expect(char *str, FILE *fd)
{
	char *s;

	if (last_message==NULL)
		last_message=stralloc(1024);
	if (buf==NULL)
		buf=stralloc(1024);

	s=fgets(buf,1024,fd);
	if (s!=NULL) {
		while ((*(s+strlen(s)-1)=='\n') || (*(s+strlen(s)-1)=='\r'))
			*(s+strlen(s)-1)='\0';
		last_message=strncpy(last_message,s,1024);

		if (strstr(s,str)!=s)
			return FALSE;
		else
			return TRUE;
	} else
		return FALSE;
}

/***
 * smtp_connect
 * Connect to the given SMTP server.
 */
FILE *smtp_connect(const char *server, int port)
{
	struct hostent *h;
	struct sockaddr_in sock;
	int sid;
	FILE *fd;
	char *c;

	h=gethostbyname(server);
	if (h==NULL) {
		log_error("smtp_connect (gethostbyname): %s\n",hstrerror(h_errno));
		return NULL;
	}

	if ((sid=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP))==-1) {
		log_error("smtp_connect (socket): %s\n",strerror(errno));
		return NULL;
	}

	sock.sin_family=AF_INET;
	sock.sin_port=htons(port);
	memcpy(&sock.sin_addr,h->h_addr,h->h_length);
	
	if (connect(sid,(struct sockaddr *) &sock,sizeof(sock))==-1) {
		log_error("smtp_connect (connect): %s\n",strerror(errno));
		return NULL;
	}

	fd=fdopen(sid,"a+");

	if (!smtp_expect("220",fd))
		return NULL;

	/*
	 * HELO name is user-configurable, with the default being the host's
	 * canonical name.
	 */
	if (conf.host_name!=NULL) {
		c=conf.host_name;
	} else {
	        c=fqdn;
	}
	if (c==NULL) {
		c="localhost";
	}
	fprintf(fd,"HELO %s\r\n",c);

	if (!smtp_expect("250",fd))
		return NULL;

	return fd;
}

/***
 * smtp_send
 * Send a message to the SMTP server.
 */
int smtp_send(t_message *message, FILE *fd, void (*status)(t_line *line))
{
	int last_was_header;
	t_line *line;
	t_address *address;

	log_enter();
	fprintf(fd,"RSET\r\n");
	smtp_expect("250",fd);

	if (message->from->formatted==NULL)
		address_format(message->from);
	fprintf(fd,"MAIL FROM:<%s>\r\n",message->from->formatted);
	if (!smtp_expect("250",fd))
		return FALSE;

	address=message->to;
	while (address!=NULL) {
		if (address->formatted==NULL)
			address=address_format(address);

		fprintf(fd,"RCPT TO:<%s>\r\n",address->formatted);
		if (!smtp_expect("250",fd))
			return FALSE;

		address=address->next;
	}

	fprintf(fd,"DATA\r\n");
	if (!smtp_expect("354",fd))
		return FALSE;

	last_was_header=TRUE;
	line=message->body;
	while (line!=NULL) {
		if (line->is_header!=last_was_header) {
			last_was_header=FALSE;
			fprintf(fd,"\r\n");
		}
		if (*(line->str)=='.')
			fprintf(fd,".");
		fprintf(fd,"%s\r\n",line->str);
		if (status!=NULL)
			status(line);
		line=line->next;
	}

	fprintf(fd,".\r\n");
	if (!smtp_expect("250",fd))
		return FALSE;

	log_leave();

	return TRUE;
}

/***
 * smtp_close
 * Close the connection to the SMTP server.
 */
void smtp_close(FILE *fd)
{
	fprintf(fd,"QUIT\r\n");
	if (!smtp_expect("221",fd)) {
		if (shutdown(fileno(fd),0)==-1)
			log_error("smtp_close (shutdown): %s\n",strerror(errno));
	}

	fclose(fd);
}

/* end smtp.c */
