/*
	smtpserv.c -- a small SMTP server for OMTA
	(C) Wouter Coene, 1999

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program. If not, write to the Free Software
	Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define WANT_SYS_TYPES
#include "config-wrap.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "smtpserv.h"
#include "message.h"
#include "mailer.h"
#include "omta.h"
#include "header.h"
#include "address.h"
#include "readconf.h"
#include "lib.h"
#include "log.h"

/***
 * smtp_get_hostname
 */
static void smtp_get_hostname(unsigned long int remote_host, char **host_name, char **host_ip)
{
	struct hostent *host_ent;
	struct in_addr in;

	in.s_addr=remote_host;
	if (remote_host==0) {
		*host_ip=strdup("127.0.0.1");
		*host_name=strdup("stdin");
	} else if (inet_netof(in)==IN_LOOPBACKNET) {
		*host_ip=strdup("127.0.0.1");
		*host_name=strdup("localhost");
	} else {
		*host_ip=strncpy(stralloc(16),inet_ntoa(in),15);
		if ((host_ent=gethostbyname(*host_ip))==NULL) {
			*host_name=stralloc(18);
			*host_name=strcat(*host_name,"[");
			*host_name=strcat(*host_name,*host_ip);
			*host_name=strcat(*host_name,"]");
		} else {
			*host_name=strdup(host_ent->h_name);
		}
	}
}

/***
 * smtp_check_relay
 * Creates an t_address struct if relaying is permitted, or returns NULL if
 * it's not.
 */
static t_address *smtp_check_relay(unsigned long int remote_host, const char *host_name,
	const char *host_ip, const char *to)
{
	t_address *address;
	struct in_addr in;
	int i;

	address=address_alloc(to);

	/*
		Allow if the message was received from stdin or from
		127.x.x.x or if it's directed to us.
	*/
	in.s_addr=remote_host;
	if ((remote_host==0)
			|| (inet_netof(in)==IN_LOOPBACKNET)
			|| (address_islocal(address,conf.local,conf.local_count)))
		return address;

	/*
		Allow everybody specified in 'RelayFor' or 'Local' lines.

		Use a static buffer here to work around libraries that use
		the same buffer for inet_ntoa() and gethostbyname()
								-- Wouter
	*/

	for (i=0; i<conf.local_count; i++) {
		if ((strcasecmp(host_name,conf.local[i])==0)
				|| (strcmp(host_ip,conf.local[i])==0))
			return address;
        }
	for (i=0; i<conf.relay_for_count; i++) {
		if ((strcasecmp(host_name,conf.relay_for[i])==0)
				|| (strcmp(host_ip,conf.relay_for[i])==0))
			return address;
        }

	/* nope, you fail */
	address_free(address);

	return NULL;
}

static t_smtp_error smtp_error[]={
	{ 214,	0,	OMTA_URI },
	{ 221,	0,	"*tsshk* Transmission ends." },
	{ 250,	0,	"Ok." },
	{ 500,	0,	"Unrecognized command." },
	{ 501,	0,	"Syntax error in command." },
	{ 502,	0,	"Command not implemented." },
	{ 503,	0,	"Bad sequence of commands." },
	{ 550,	0,	"Information unavailable or access denied." },
	{ 551,	0,	"Relay denied due to policy regulations." },
	{ 552,	0,	"Too much mail data." },
	{ 554,	0,	"Transaction failed." },
	{ 1000,	250,	"Hi." },
	{ 2000,	503,	"Polite people say HELO first." },
	{ 1001,	250,	"Sender accepted." },
	{ 2001,	503,	"Already did this command." },
	{ 1002,	250,	"Recipient accepted." },
	{ 2002,	503,	"Need a valid sender first" },
	{ 1003,	250,	"Message accepted." },
	{ 2003,	503,	"Need valid sender and recipients." },
	{ 0,	0,	"Unknown internal error." }
};

/***
 * smtp_reply
 * Print the human readable message along with the given SMTP return code.
 */
static void smtp_reply(int ret, FILE *fd, t_smtp_data *data)
{
	int i;

	i=0;
	forever {
		if ((smtp_error[i].error==ret) || (smtp_error[i].error==0)) {
			if (smtp_error[i].xlat>0)
				ret=smtp_error[i].xlat;
			log_dialog("%s: %s%s%s -> %d %s\n",data->host_name,data->cmd,
				data->args!=NULL? " ":"",data->args!=NULL? data->args:"",ret,
				smtp_error[i].message);
			fprintf(fd,"%d %s\r\n",ret,smtp_error[i].message);
			break;
		}
		i++;
	}
}

/***
 * smtp_helo
 * Implements HELO.
 */
static int smtp_helo(const char *args, FILE *fdin, FILE *fdout, t_smtp_data *data)
{
	smtp_reply(1000,fdout,data);
	message_set_helo(data->message,args);

	return -1;
}

/***
 * smtp_mail
 * Implements MAIL.
 */
static int smtp_mail(const char *args, FILE *fdin, FILE *fdout, t_smtp_data *data)
{
	if (data->did_from)
		return 2001;

	message_set_sender(data->message,address_alloc(args));
	data->did_from=TRUE;

	return 1001;
}

/***
 * smtp_rcpt
 * Implements RCPT.
 */
static int smtp_rcpt(const char *args, FILE *fdin, FILE *fdout, t_smtp_data *data)
{
	t_address *address;

	if (!data->did_from)
		return 2002;

	if ((address=smtp_check_relay(data->remote_host,data->host_name,data->host_ip,args))==NULL) {
		return 551;
	}

	message_add_recipient(data->message,address);
	data->did_to=TRUE;

	return 1002;
}

/***
 * smtp_data
 * Implements DATA.
 */
static int smtp_data(const char *args, FILE *fdin, FILE *fdout, t_smtp_data *data)
{
	if ((!data->did_from) || (!data->did_to))
		return 2003;

	fprintf(fdout,"354 Proceed. End your message with a single dot ('.') on a line.\r\n");
	mailer_get_message(data->message,fdin,FALSE,TRUE);
	if (data->message->error!=MSGERROR_OK) {
		smtp_reply(data->message->error,fdout,data);
	} else {
		smtp_reply(mailer_send(data->message)? 1003:554,fdout,data);
	}

	return -3;
}

static t_smtp_command smtp_command[]={
	{ "HELO",	smtp_helo,	FALSE,	TRUE,	FALSE,	0 },
	{ "HELP",	NULL,		FALSE,	FALSE,	FALSE,	214 },
	{ "MAIL",	smtp_mail,	TRUE,	TRUE,	TRUE,	0 },
	{ "RCPT",	smtp_rcpt,	TRUE,	TRUE,	TRUE,	0 },
	{ "DATA",	smtp_data,	FALSE,	FALSE,	TRUE,	0 },
	{ "RSET",	NULL,		FALSE,	FALSE,	TRUE,	-2 },
	{ "QUIT",	NULL,		FALSE,	FALSE,	FALSE,	-4 },
	{ "NOOP",	NULL,		FALSE,	FALSE,	FALSE,	250 },
	{ "VRFY",	NULL,		FALSE,	TRUE,	TRUE,	550 },
	{ "EXPN",	NULL,		FALSE,	TRUE,	TRUE,	550 },
	{ NULL,		NULL,		FALSE,	FALSE,	FALSE,	0 }
};

/***
 * smtp_server
 * Run a SMTP server on the given streams.
 */
int smtp_server(FILE *fdin, FILE *fdout, const char *greeting,
	unsigned long int remote_host)
{
	char *buf, *s;
	char *cmd, *args, *ucase_args, *helo;
	int i, ret;
	t_smtp_data data;

	data.message=message_alloc();
	data.message->from_host.s_addr=remote_host;
	data.did_from=FALSE;
	data.did_to=FALSE;
	data.remote_host=remote_host;

	smtp_get_hostname(remote_host,&(data.host_name),&(data.host_ip));
	if (remote_host!=0)
		log_status("Connection from %s (%s)\n",data.host_name,data.host_ip);

	buf=stralloc(MAX_LINE_LEN);
	helo=NULL;

	fprintf(fdout,"220 %s\r\n",greeting);
	forever {
		s=xgetline(buf,MAX_LINE_LEN,fdin);
		if (feof(fdin) || (s==NULL))
			break;

		cmd=trim(s," \t\r\n");

		args=strchr(s,' ');
		if (args!=NULL) {
			*args='\0';
			args++;
		}
		cmd=strtoupper(cmd);
		if (args!=NULL)
			ucase_args=strtoupper(strdup(args));
		else
			ucase_args=NULL;

		data.cmd=cmd;
		data.args=args;
		i=0;
		ret=500;
		forever {
			if (smtp_command[i].command==NULL)
				break;

			if (strcmp(cmd,smtp_command[i].command)==0) {
				if ((smtp_command[i].has_from_to) && (args!=NULL)) {
					if (strstr(ucase_args,"FROM:")==ucase_args) {
						args+=5;
						args=trim(args," \t");
					} else if (strstr(ucase_args,"TO:")==ucase_args) {
						args+=3;
						args=trim(args," \t");
					}
				}
				if ((smtp_command[i].post_helo) && (helo==NULL)) {
					ret=503;
				} else if (smtp_command[i].need_args==(args==NULL)) {
					ret=501;
				} else {
					if (smtp_command[i].handler!=NULL)
						ret=smtp_command[i].handler(args,fdin,fdout,&data);
					else if (smtp_command[i].ret!=0)
						ret=smtp_command[i].ret;
					else
						ret=502;
				}
				break;
			}

			i++;
		}

		if (ucase_args!=NULL)
			free(ucase_args);

		if (ret>0) {
			smtp_reply(ret,fdout,&data);
		} else if (ret==-1) {
			helo=strdup(data.message->helo);
		} else if ((ret==-2) || (ret==-3)) {
			message_free(data.message);
			data.message=message_alloc();
			data.message->from_host.s_addr=remote_host;
			message_set_helo(data.message,helo);

			data.did_from=FALSE;
			data.did_to=FALSE;

			if (ret==-2)
				smtp_reply(250,fdout,&data);
		} else {
			smtp_reply(221,fdout,&data);
			break;
		}
	}

	if (remote_host!=0)
		log_status("Closed connection to %s (%s)\n",data.host_name,data.host_ip);

	free(buf);
	message_free(data.message);

	return TRUE;
}

/* end smtpserv.c */
